@@ -1,11 +1,13 @@
-__all__ = ['ResultException', 'ResultType', 'Result', 'Ok', 'Err', 'wrap_result', 'auto_unwrap']
+__all__ = [
+    'ResultException', 'ResultType', 'Result', 'Ok', 'Err',
+    'WrappedBase', 'UnwrappedBase', 'wrap_result', 'auto_unwrap',
+]
 
 import types
 import asyncio
 import inspect
+from typing import Any, Callable, ForwardRef
 from functools import wraps
-from typing import Any, Callable
-from inspect import iscoroutinefunction
 
 
 class ResultException(Exception):
@@ -77,15 +79,22 @@ class ResultType(_ResultType):
             E = cls.__annotations__['e']
             
             if asyncio.iscoroutinefunction(fn):
-                # @wraps(fn)
+                @wraps(fn)
                 async def wrap(*args, **kwargs) -> cls:
                     try:
                         v = await fn(*args, **kwargs)
                         return Ok[V](v)
                     except Exception as e:
                         return Err[E](e)
+            elif inspect.isfunction(fn) or inspect.ismethod(fn):
+                @wraps(fn)
+                def wrap(*args, **kwargs) -> cls:
+                    try:
+                        v = fn(*args, **kwargs)
+                        return Ok[V](v)
+                    except Exception as e:
+                        return Err[E](e)
             elif callable(fn):
-                # @wraps(fn)
                 def wrap(*args, **kwargs) -> cls:
                     try:
                         v = fn(*args, **kwargs)
@@ -137,10 +146,11 @@ class Ok(Result):
     def __init__(self, v: Any):
         V: type = self.__class__.__annotations__['v']
 
-        if not (isinstance(V, types.GenericAlias)
-                or isinstance(V, str)
-                or V is Any
-                or isinstance(v, V)):
+        if not (isinstance(V, types.GenericAlias) or
+                isinstance(V, str) or
+                V is Any or
+                isinstance(V, ForwardRef) or
+                isinstance(v, V)):
             raise TypeError(f'Got {type(v)} but expected {self.__class__.__annotations__["v"]}')
         
         self.v = v
@@ -180,10 +190,11 @@ class Err(Result):
     def __init__(self, e: Any):
         E: type = self.__class__.__annotations__['e']
 
-        if not (isinstance(E, types.GenericAlias)
-                or isinstance(E, str)
-                or E is Any
-                or isinstance(e, (E, Exception))):
+        if not (isinstance(E, types.GenericAlias) or
+                isinstance(E, str) or
+                E is Any or
+                isinstance(E, ForwardRef) or
+                isinstance(e, (E, Exception))):
             raise TypeError(f'Got {type(e)} but expected {self.__class__.__annotations__["e"]}')
 
         self.e = e
@@ -222,7 +233,7 @@ def wrap_result(res: Result):
     def outer(f):
         @wraps(f)
         def inner(*args, **kwargs):
-            if iscoroutinefunction(f):
+            if asyncio.iscoroutinefunction(f):
                 async def a():
                     try:
                         v = await f(*args, **kwargs)
@@ -243,50 +254,138 @@ def wrap_result(res: Result):
     return outer
 
 
-def auto_unwrap(obj) -> Any:
-    """
-    This function auto unwraps Ok or Err type of Result
-    """
-    if inspect.isclass(obj):
-        type_ = obj
-        new_type_dict = {}
+class WrappedBase:
+    def __getstate__(self) -> dict:
+        state = {}
 
-        for a in dir(type_):
-            f = getattr(type_, a)
-
-            if not callable(f):
+        for a in dir(self):
+            if a.startswith('__') and a.endswith('__'):
                 continue
 
-            if not hasattr(f, '__annotations__'):
+            v = getattr(self, a)
+
+            if callable(v):
                 continue
 
-            return_type = f.__annotations__.get('return')
+            state[a] = v
 
-            if not return_type:
-                continue  # pragma: no cover
-            
-            if not isinstance(return_type, type):
-                continue
+        return state
 
-            if not issubclass(return_type, Result):
-                continue
 
-            new_f = auto_unwrap(f)
-            new_type_dict[a] = new_f
+class UnwrappedBase:
+    def __setstate__(self, state):
+        for a, v in state.items():
+            setattr(self, a, v)
 
-        new_type = type(type_.__name__, (type_,), new_type_dict)
-        return new_type
-    else:
-        func = obj
+
+def _auto_unwrap_type(type_):
+    new_type_dict = {}
+    new_type = type(type_.__name__, (type_, UnwrappedBase), new_type_dict)
+
+    for a in dir(type_):
+        f = getattr(type_, a)
+
+        if not callable(f):
+            # print('not callable', a, f)
+            continue
+
+        if not hasattr(f, '__annotations__'):
+            # print('no __annotations__', a, f)
+            continue
+
+        return_type = f.__annotations__.get('return')
+
+        if not return_type:
+            # print('no return_type', a, f)
+            continue  # pragma: no cover
         
-        @wraps(func)
-        def wrapper(*args, **kwargs):
-            if iscoroutinefunction(obj):
-                async def a():
-                    return (await func(*args, **kwargs)).unwrap()
+        '''
+        if not (isinstance(return_type, type) or isinstance(return_type, str)):
+            print('return_type is not type/str', a, f, return_type, type(return_type))
+            continue
+
+        if not (issubclass(return_type, Result) or isinstance(return_type, str)):
+            print('return_type is Result/str subclass', a, f, return_type, type(return_type))
+            continue
+        '''
+
+        '''
+        if isinstance(return_type, type) and issubclass(return_type, Result):
+            pass
+        elif isinstance(return_type, str):
+            pass
+        else:
+            print('return_type is not type Result/str', a, f, return_type, type(return_type))
+            continue
+        '''
+
+        new_f = _auto_unwrap_func(f, old_type=return_type, new_type=new_type)
+        setattr(new_type, a, new_f)
+    
+    return new_type
+
+
+def _auto_unwrap_func(func, old_type=None, new_type=None):
+    # print('auto_unwrap func-like', func, old_type, new_type)
+
+    @wraps(func)
+    def wrapper(*args, **kwargs):
+        if asyncio.iscoroutinefunction(func):
+            async def a():
+                r = await func(*args, **kwargs)
+
+                if isinstance(r, Result):
+                    r = r.unwrap()
                 
-                return a()
-            else:
-                return func(*args, **kwargs).unwrap()
+                if (inspect.isclass(old_type) and issubclass(old_type, WrappedBase) and isinstance(r, old_type)) or \
+                   (isinstance(old_type, ForwardRef) and old_type.__forward_arg__ == r.__class__.__name__):
+                    new_r = new_type.__new__(new_type)
+                    state = r.__getstate__()
+                    new_r.__setstate__(state)
+                    r = new_r
+                
+                return r
+            
+            return a()
+        elif inspect.ismethod(func):
+            r = func(*args, **kwargs)
+
+            if isinstance(r, Result):
+                r = r.unwrap()
+            
+            if (inspect.isclass(old_type) and issubclass(old_type, WrappedBase) and isinstance(r, old_type)) or \
+               (isinstance(old_type, ForwardRef) and old_type.__forward_arg__ == r.__class__.__name__):
+                new_r = new_type.__new__(new_type)
+                state = r.__getstate__()
+                new_r.__setstate__(state)
+                r = new_r
+
+            return r
+        else:
+            r = func(*args, **kwargs)
+
+            if isinstance(r, Result):
+                r = r.unwrap()
+            
+            if (inspect.isclass(old_type) and issubclass(old_type, WrappedBase) and isinstance(r, old_type)) or \
+               (isinstance(old_type, ForwardRef) and old_type.__forward_arg__ == r.__class__.__name__):
+                new_r = new_type.__new__(new_type)
+                state = r.__getstate__()
+                new_r.__setstate__(state)
+                r = new_r
+
+            return r
 
     return wrapper
+
+
+def auto_unwrap(obj) -> Any:
+    '''
+    This function auto unwraps Ok or Err type of Result
+    '''
+    if inspect.isclass(obj):
+        r = _auto_unwrap_type(obj)
+    else:
+        r = _auto_unwrap_func(obj)
+    
+    return r
